#!/usr/bin/env python
# -*- coding: UTF-8 -*-


from artix.common import py2to3utils
from propertyReader import properties
import json
import logger
import os
import pickle
import queuesManager as queuesManager
import sendSalesModule as sendSalesModule
import socket
import sys
import threading
import time
import traceback
import uuid
import zipfile

# импорт необходим, в нем молча происходят некоторые подготовительные работы для того, чтоб логгер начал работать
from plugins import backupFile, execscript, status
from syncThrift import CashCommunicationService
from syncThrift.ttypes import *
from thrift.server import TServer
from thrift.transport import TSocket

loggerSync = logger.getLogger()

__THRIFT_LISTENING_PORT = 7795
__THRIFT_LISTENING_HOST = '0.0.0.0'


def qpidTasksHandle():
    if queuesManager.checkUnloadSalesOnSkipMode() == status.skipped:
        # молча ждем внешнего события, которое сбросит ошибочное состояние
        return queuesManager.__TASK_HANDLE_LONG_SLEEP_TIME

    defferedSalesUnloadResult = queuesManager.checkUnloadedSalesQpid()
    if defferedSalesUnloadResult != status.OK:
        queuesManager.updateUnloadedSalesTaskState(True)
        return queuesManager.__TASK_HANDLE_LONG_SLEEP_TIME

    nextSleep, res = queuesManager.handleTasks(queuesManager.THROUGH_QPID)
    return nextSleep


def grpcTasksHandle():
    if queuesManager.checkUnloadSalesOnSkipMode() == status.skipped:
        # молча ждем внешнего события, которое сбросит ошибочное состояние
        return queuesManager.__TASK_HANDLE_LONG_SLEEP_TIME

    nextSleep, res = queuesManager.handleTasks(queuesManager.THROUGH_GRPC)

    unloadStatus, successUnloadedSalesIds, messEn, messRu = sendSalesModule.sendGrpcUnloadedSales()
    queuesManager.removeUnloadedSalesByGrpc(successUnloadedSalesIds)
    if unloadStatus not in [status.OK, status.empty]:
        nextSleep = queuesManager.__TASK_HANDLE_LONG_SLEEP_TIME
        # Если по какой-либо причине не удалось отправить сообщение на сервер - оно не отправится до тех пор, пока не сбросится ошибочное состояние.
        # Сбросить его можно при перезапуске sync-а или при получении нового запроса на выгрузку продаж через thrift
        tasks = queuesManager.updateUnloadedSalesTaskState(True)
        try:
            for task in tasks:
                if task.get('saleStateId'):
                    status.updateSaleState(task.get('saleStateId'), messEn, messRu, status.STATUS_ERROR,
                            properties.httpHostCS, properties.httpPortCS)
        except Exception as e:
            loggerSync.warn("Exception while send status on server: %s", e.message)
            loggerSync.warn(traceback.format_exc())
    return nextSleep


def unloadTasksHandle(threadHandlerController):
    while threadHandlerController.get("continueTaskHandler"):
        try:
            if properties.protocolVersion >= 4:
                nextSleep = grpcTasksHandle()
            else:
                nextSleep = qpidTasksHandle()
        except Exception:
            nextSleep = queuesManager.__TASK_HANDLE_LONG_SLEEP_TIME
            loggerSync.error(
                'Произошла ошибка во время обработки задач на выгрузку. Попробуем обработать через минуту.\n{0}'
                .format(traceback.format_exc())
                )
        loggerSync.debug("Sleep: %s", nextSleep)
        time.sleep(nextSleep)


def removeRoutesByQueue(broker, queue, remoteBroker, routeDestExchange):
    from qmf.console import Session
    sess = Session()
    _connTimeout = 10
    localBroker = sess.addBroker(broker, _connTimeout)
    localBroker._waitForStable()
    localBroker.getAmqpSession().queue_declare(queue=queue, passive=False, durable=True, arguments={'qpid.file_size': 24, 'qpid.file_count': 8})
    bridges = sess.getObjects(_class="bridge", dynamic=False)
    allLinks = []
    for bridge in bridges:
        allLinks.append(bridge._linkRef_)
    changedRemoteBroker = True
    for bridge in bridges:
        if bridge.src == queue:
            remotelink = bridge._linkRef_
            if remoteBroker.split(':')[0] == remotelink.host and remoteBroker.split(':')[1] == str(remotelink.port) and bridge.dest == routeDestExchange and bridge.sync:
                changedRemoteBroker = False
                loggerSync.info('Routing already added')
                continue

            for i in range(len(allLinks)):
                if allLinks[i].host == remotelink.host and allLinks[i].port == remotelink.port:
                    allLinks.__delitem__(i)
                    break
            loggerSync.info("Removing route (from broker=%s queue=%s to broker=%s:%s exchange=%s)" % (
                bridge.getBroker().getUrl(), bridge.src, remotelink.host, remotelink.port, bridge.dest))
            res = bridge.close()
            if res.status != 0:
                loggerSync.error("Error closing bridge: %d - %s" % (res.status, res.text))

            delLinkFlag = True
            for i in range(len(allLinks)):
                if allLinks[i].host == remotelink.host and allLinks[i].port == remotelink.port:
                    delLinkFlag = False
                    break
            if delLinkFlag:
                loggerSync.info("Closing link %s:%s" % (remotelink.host, remotelink.port))
                res = remotelink.close()
                if res.status != 0:
                    loggerSync.error("Error closing link: %d - %s" % (res.status, res.text))

    try:
        sess.delBroker(localBroker)
        sess.close()
    except Exception as e:
        loggerSync.error('Error on broker session closing. %s' % str(e))
    return changedRemoteBroker


class CashCommunicationHandler:
    lockLoad = threading.Lock()
    lockExecute = threading.Lock()

    def load(self, loadScript, configInfo=None):
        method = 'load'
        res = status.errorProgramm
        if self.lockLoad.acquire(False):
            try:
                reconfigure(configInfo.__dict__ if configInfo else None)
                status.sendStatus(loggerSync, method, status.start, 'Start load', properties.httpHostCS, properties.httpPortCS)
                loggerSync.info('Начинаем загружать справочники')

                loggerSync.info("Идентификатор потока, захватившего загрузку %s, %s" % (str(threading.current_thread().ident), threading.current_thread().name))
                pluginName = 'sync' if configInfo and configInfo.__dict__.get('protocolVersion', 4) < 5 else 'sync_rest'
                plugin = __import__(pluginName, fromlist='q')
                rez, mess = plugin.main(properties)
                loggerSync.info('Вернулись из плагина')
                status.sendStatus(loggerSync, method, rez, mess, properties.httpHostCS, properties.httpPortCS)
                loggerSync.info('Отправили статус на сервер')
            except Exception as e:
                loggerSync.error('Error 0, %s' % str(e))
                status.sendStatus(loggerSync, method, status.errorProgramm, str(e), properties.httpHostCS, properties.httpPortCS)
                res = status.errorProgramm
            finally:
                loggerSync.info('Зашли в finally')
                loggerSync.info("Идентификатор потока, разблокировавший загрузку %s, %s"
                        % (str(threading.current_thread().ident), threading.current_thread().name))
                self.lockLoad.release()
                loggerSync.info('Процесс загрузки разблокирован')
                return res
        else:
            loggerSync.info('Процесс загрузки справочников занят!')
            loggerSync.info("Отказано потоку с идентификатором %s, %s"
                    % (str(threading.current_thread().ident), threading.current_thread().name))
            status.sendStatus(loggerSync, method, status.busy, 'Busy', properties.httpHostCS, properties.httpPortCS)
            return status.busy

    def unload(self, unloadScript, fromDate, toDate, shopId=None, cashCode=None, configInfo=None):
        method = 'unload'
        configInfoDict = configInfo.__dict__ if configInfo else None
        saleStateId = None
        try:
            reconfigure(configInfoDict)
            httpHostCS = properties.httpHostCS
            httpPortCS = properties.httpPortCS
            queuesManager.updateUnloadedSalesTaskState(False)
            addToQueue = True
            # если в очереди уже есть запросы с диапазоном дат входящим в диапазон текущего запроса, то изменим начало текущего запроса
            try:
                with queuesManager.unloadTasksLocker:
                    queueTasks = queuesManager.getQueueByName(queuesManager.unloadTasksQueueName)
                    addToQueue, fromDate, saleStateId = self.normalizeTasks(fromDate, toDate, shopId, cashCode,
                            saleStateId, queueTasks, httpHostCS, httpPortCS)
                    queueTasks.manager.close()
                if addToQueue:
                    with queuesManager.unloadedSalesTasksLocker:
                        unloadedSalesQueue = queuesManager.getQueueByName(queuesManager.unloadedSalesTasksQueueName)
                        addToQueue, fromDate, saleStateId = self.normalizeTasks(fromDate, toDate, shopId, cashCode,
                                saleStateId, unloadedSalesQueue, httpHostCS, httpPortCS)
                        unloadedSalesQueue.manager.close()
            except Exception as e:
                loggerSync.warn(u'Не удалось проверить необработанные запросы продаж. %s' % str(e))
                loggerSync.error(traceback.format_exc())
            if addToQueue:
                saleStateId = status.addSaleState('Putting task for unload sales to queue', 'Ставим в очередь задачу на выгрузку продаж',
                                                  status.STATUS_DURING,
                                                  httpHostCS, httpPortCS, fromDate, toDate, callFromCS=True)
                task = {'fromDate': fromDate,
                        'toDate': toDate,
                        'shopId': shopId,
                        'cashCode': cashCode,
                        'configInfo': configInfoDict,  # преобразованный в словарь набор параметров, которые прислал КС. Вопрос, нужны ли они здесь.
                        'saleStateId': saleStateId,
                        'httpHostCS': httpHostCS,
                        'httpPortCS': httpPortCS,
                        'taskId': str(uuid.uuid4()) # для того, чтоб была возможность связать задачу на выгрузку с выгруженной продажей, генерирую идентификатор к ней уже здесь
                        }
                with queuesManager.unloadTasksLocker:
                    queueTasks = queuesManager.getQueueByName(queuesManager.unloadTasksQueueName)
                    queueTasks.push(queuesManager.Message(json.dumps(task).encode('utf-8'), ttl=60, flags=queuesManager.FLAG_PERSISTENT))
                    queueTasks.manager.close()
                loggerSync.info(
                    u'Задача на выгрузку продаж поставлена в очередь (fromDate=%s, toDate=%s, shopId=%s, cashCode=%s).' % (
                        fromDate, toDate, shopId, cashCode))
        except Exception as e:
            loggerSync.error("Error {0}\n{1}\n{2}".format(e.__class__.__name__, str(e), traceback.format_exc()))
            status.sendStatus(loggerSync, method, status.errorProgramm, str(e),
                    properties.httpHostCS, properties.httpPortCS)
            status.updateSaleState(saleStateId, str(e), str(e), status.STATUS_ERROR,
                    properties.httpHostCS, properties.httpPortCS)
            return status.errorProgramm
        finally:
            loggerSync.info('Зашли в finally')
            loggerSync.info('Процесс выгрузки разблокирован')
        return status.OK

    def unloadSalesFile(self, saleFileZip, salesProperty):
        """
        Метод используется сервисом artixcs-sales-loader, чтобы пришедшие продажи на МКС через GRPC переслать на ЦКС.
        saleFileZip - zip-архив файла с продажами. Это тот же самый архив, что sync-агент отправляет по GRPC.
        """
        method = 'unloadSalesFile'
        httpHostCS = properties.httpHostCS
        httpPortCS = properties.httpPortCS

        try:
            queuesManager.updateUnloadedSalesTaskState(False)
            loggerSync.info("Формирую свойства для повторной отправки продаж: ")
            prop = {'cashCode': salesProperty.cashId,
                    # значение уходит в qpid, а там читается на стороне КС, и в нем ожидается именно ID кассы, поэтому такая путаница
                    # с названием кода. Хотя, конечно, причем здесь qpid, когда сообщение должно слаться через grpc? - отложенные продажи
                    # могут внезапно пойти через qpid
                    'shopId': salesProperty.shopId,
                    'shopCode': salesProperty.shopCode,
                    'realCashCode': salesProperty.cashCode,  # тот cashCode, который везде как cashCode, а здесь поле с таким именем уже занято.
                    'compressMethod': 'zip',
                    'cashUnloadState': salesProperty.cashUnloadState,
                    'messageType': 'sale',
                    'httpHostCS': httpHostCS,
                    'httpPortCS': httpPortCS,
                    'salesId': str(uuid.uuid4())}
            if salesProperty.dateTimeBegin:
                prop.update({'saleType': 'dates', 'CSReguestBegin': salesProperty.dateTimeBegin, 'CSReguestEnd': salesProperty.dateTimeEnd})
            elif salesProperty.shiftNumberBegin:
                prop.update({'saleType': 'shifts', 'CSReguestBeginShift': salesProperty.shiftNumberBegin, 'CSReguestEndShift': salesProperty.shiftNumberEnd})

            loggerSync.info("Свойства для повторной отправки продаж сформированы: ")
            loggerSync.info(prop)

            pathSales = backupFile.getUnloadedSalesPath()
            loggerSync.info("Архивирую продажи для последующей отправки")
            propFileName = 'sale.properties'
            pathTmpRouting = backupFile.prepareDirectory('tmpRouting')
            pathRoutingSales = '%s/routing_sales%s.zip' % (pathTmpRouting, prop.get('salesId'))

            with open(pathRoutingSales, 'wb') as archive:
                archive.write(saleFileZip)

            archive = zipfile.ZipFile(pathRoutingSales, 'a', 8, True)
            with open(propFileName, 'wb') as propFile:
                pickle.dump(prop, propFile, pickle.HIGHEST_PROTOCOL)
            archive.write(propFileName, None, 8)
            archive.close()
            backupFile.remove(propFileName)
            backupFile.move(pathRoutingSales, pathSales)
            loggerSync.info("Продажи помещены в архив")
        except Exception as e:
            loggerSync.error("Error {0}\n{1}\n{2}".format(e.__class__.__name__, str(e), traceback.format_exc()))
            status.sendStatus(loggerSync, method, status.errorProgramm, str(e), httpHostCS, httpPortCS)
            return status.errorProgramm
        finally:
            loggerSync.info('Зашли в finally')
            loggerSync.info('Процесс выгрузки разблокирован')
        return status.OK

    def execute(self, command, isNewThread):
        loggerSync.info('С КС получен скрипт для запуска')
        if self.lockExecute.acquire(False):
            if isNewThread:
                try:
                    threading.Thread(target=execscript.execute, args=(command,)).start()
                    return ExecuteResult(out='Скрипт успешно запущен', err='')
                except Exception as e:
                    loggerSync.error(str(e))
                    return ExecuteResult(out='', err='%s' % str(e))
                finally:
                    self.lockExecute.release()
            else:
                try:
                    result = execscript.execute(command)
                    return ExecuteResult(out=result.get('out'), err=result.get('err'))
                except Exception as e:
                    loggerSync.error(str(e))
                    return ExecuteResult(out='', err='%s' % str(e))
                finally:
                    self.lockExecute.release()
        loggerSync.warn('Процесс занят')
        return ExecuteResult(out='', err='Процесс занят')

    def normalizeTasks(self, fromDate, toDate, shopId, cashCode, saleStateId, queueTasks, host, port):
        addToQueue = True
        for i in range(len(queueTasks)):
            taskStr = queueTasks.queue[i].data
            if taskStr != None:
                taskFromQueue = json.loads(py2to3utils.ensureStr(taskStr))
                if taskFromQueue['shopId'] == shopId and taskFromQueue['cashCode'] == cashCode:
                    if taskFromQueue['fromDate'] == fromDate and taskFromQueue['toDate'] == toDate:
                        messRu = 'Задача за диапазон [%s - %s] уже поставлена в очередь на обработку' % (fromDate, toDate)
                        loggerSync.info(messRu)
                        addToQueue = False
                        saleStateId = status.addSaleState('Task for date [%s - %s] already in queue' % (fromDate, toDate), messRu, status.STATUS_READY,
                                                          host, port, fromDate, toDate, callFromCS=True)
                    elif taskFromQueue['fromDate'] <= fromDate and taskFromQueue['toDate'] > fromDate and taskFromQueue['toDate'] < toDate:
                        loggerSync.info(
                            u'В очереди уже есть запрос на выгрузку продаж, диапазон дат которого перекрывает часть текущего запроса. Изменяем время начала запроса продаж %s -> %s' % (
                                fromDate, taskFromQueue['toDate']))
                        fromDate = taskFromQueue['toDate']
        return addToQueue, fromDate, saleStateId

    # этот метод  вызывается при активации МКС, чтобы настроить подключение к ЦКС
    def reconfigureSyncProperties(self, configInfo):
        loggerSync.info(u'Вызвана перезапись настроек sync')
        reconfigure(configInfo.__dict__ if configInfo else None)


def reconfigure(configInfo):
    if properties.update(configInfo) and properties.protocolVersion < 4:
        configureRouting()


def createAndBindQueueToExchane(broker, queue, exchange=None, routingKey='#'):
    if exchange:
        from qmf.console import Session
        sess = Session()
        _connTimeout = 10
        localBroker = sess.addBroker(broker, _connTimeout)
        localBroker._waitForStable()
        localBroker.getAmqpSession().queue_declare(queue=queue, passive=False, durable=True, arguments={'qpid.file_size': 24, 'qpid.file_count': 8})

        localBroker.getAmqpSession().exchange_declare(exchange=exchange, passive=False, durable=True, type='topic', auto_delete=False)
        localBroker.getAmqpSession().exchange_bind(exchange=exchange, queue=queue, binding_key=routingKey)
        try:
            sess.delBroker(localBroker)
            sess.close()
        except Exception as e:
            loggerSync.error('Error on broker session closing. %s' % str(e))


def configureRouting():
    routeQueue = 'artix.backward.cs'
    bindExchange = routeDestExchange = 'artix.backward.cs'
    isConfigure = False
    times = 5  # количество попыток
    error = ''
    while not isConfigure and times:
        loggerSync.info('Пытаемся подключится к локальному серверу')
        times -= 1
        try:
            createAndBindQueueToExchane(broker=properties.localBroker, queue=routeQueue, exchange=bindExchange)
            if properties.mode == 'server':
                routeQueue = 'artix.backward.mcs'
                createAndBindQueueToExchane(broker=properties.localBroker, queue=routeQueue, exchange=bindExchange)
            addNewRoute = removeRoutesByQueue(broker=properties.localBroker, queue=routeQueue,
                    remoteBroker=properties.remoteBroker.split('@')[1], routeDestExchange=routeDestExchange)
            if addNewRoute:
                loggerSync.info("Adding route from broker=%s queue=%s to broker=%s exchange=%s" % (
                    properties.localBroker.split('@')[1], routeQueue, properties.remoteBroker.split('@')[1], routeDestExchange))
                os.system("qpid-route -d -s --ack=1 queue add %s %s %s %s"
                        % (properties.remoteBroker, properties.localBroker, routeDestExchange, routeQueue))
        except socket.error as e:
            error = str(e)
            time.sleep(10)
        else:
            isConfigure = True
    if not isConfigure:
        loggerSync.error('Ошибка подключения к локальному сервису qpid. %s ' % error)


#
#   Данный метод необходимо выполнить здесь, так как 
#   в exchangers в модуле startReader, который используется для формирования
#   выгрузки продаж, используется модуль signal, который должен быть подгружен
#   только в главном потоке. Иначе будет ошибка импортирования.
#   Такой же импорт находится в модуле artixinterchange.py
#
def importUnloadModule():
    try:
        curPath = sys.path[0]
        unloadScript = properties.unloadScript
        sys.path[0] = '%s' % os.path.dirname(unloadScript)
        __import__(os.path.basename(os.path.splitext(unloadScript)[0]), fromlist='q')
        sys.path[0] = curPath
    except Exception as e:
        messRu = 'Ошибка при подключении модуля выгрузки продаж.'
        loggerSync.error('%s %s' % (messRu, str(e)))
        os._exit(1)


def serveThrift(threadHandlerController):
    transport = TSocket.TServerSocket(__THRIFT_LISTENING_HOST, __THRIFT_LISTENING_PORT)
    handler = CashCommunicationHandler()
    processor = CashCommunicationService.Processor(handler)
    tfactory = TTransport.TBufferedTransportFactory()
    pfactory = TBinaryProtocol.TBinaryProtocolFactory()

    loggerSync.info('Starting the server...')

    class SyncThriftServer(TServer.TThreadedServer):
        def serve(self):
            self.serverTransport.listen()
            threadNumber = 0
            while threadHandlerController.get('continueThriftHandler'):
                threadNumber += 1
                try:
                    client = self.serverTransport.accept()
                    t = threading.Thread(target=self.handle, name="thandler_%s" % threadNumber, args=(client,))
                    t.daemon = self.daemon
                    t.start()
                except KeyboardInterrupt:
                    raise
                except Exception as x:
                    loggerSync.error(str(x))

    # You could do one of these for a multithreaded server
    server = SyncThriftServer(processor, transport, tfactory, pfactory, daemon=True)
    server.serve()
    loggerSync.info('Thrift server gone')


def runThriftThread(threadHandlerController):
    thriftThread = threading.Thread(target=serveThrift, name="thrift", args=(threadHandlerController,))
    thriftThread.daemon = True
    thriftThread.start()
    return thriftThread


def runUnloadTaskHandler(threadHandlerController):
    unloadTasksHandler = threading.Thread(target=unloadTasksHandle, name="tasks", args=(threadHandlerController,))
    unloadTasksHandler.daemon = True
    unloadTasksHandler.start()
    return unloadTasksHandler


def stackTrace():
    """
    Если вдруг sync начнет делать непонятности в рантайме, и будет полезен stack trace его работы - создайте файл
    /tmp/sync_stack_trace (touch /tmp/sync_stack_trace), и stacktrace всех потоков появится в логе. Такой себе backdoor
    :return:
    """
    while True:
        time.sleep(10)
        if os.path.exists("/tmp/sync_stack_trace"):
            code = []
            for threadId, stack in sys._current_frames().items():
                code.append("\n# ThreadID: %s" % threadId)
                for filename, lineno, name, line in traceback.extract_stack(stack):
                    code.append('File: "%s", line %d, in %s' % (filename, lineno, name))
                    if line:
                        code.append("  %s" % (line.strip()))

            loggerSync.info("Stacktrace:")
            for line in code:
                loggerSync.info(line)
            os.remove("/tmp/sync_stack_trace")


def startStackTraceKeeper():
    stackTraceThread = threading.Thread(target=stackTrace, name="stackTraceKeeper")
    stackTraceThread.daemon = True
    stackTraceThread.start()


def startSync():
    importUnloadModule()
    startStackTraceKeeper()
    if properties.protocolVersion < 4:
        configureRouting()
    else:
        loggerSync.info('Использую grpc протокол для доставки продаж на сервер. Параметры подключения: %s:%s',
                properties.grpcHostCS, properties.grpcPortCS)
    queuesManager.updateUnloadedSalesTaskState(False)
    threadHandlerController = {"continueTaskHandler": True,
                               "continueThriftHandler": True}  # Необходимо для тестов, чтоб была возможность остановить handler, если это потребуется
    runUnloadTaskHandler(threadHandlerController)
    thriftThread = runThriftThread(threadHandlerController)
    thriftThread.join()


if __name__ == '__main__':
    try:
        startSync()
    except Exception as e:
        print("Fatal exception:\n")
        traceback.print_exc()
        print(str(e))
        loggerSync.fatal("Fatal exception: %s", str(e))
        loggerSync.fatal(traceback.format_exc())
        raise e
    loggerSync.info("Shutdown sync")
